# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'Qrlew'
copyright = '2023, Sarus Technologies'
author = 'Sarus Technologies'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ['myst_parser']

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# MyST conf
myst_enable_extensions = ["colon_fence"]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'pydata_sphinx_theme'

html_static_path = ['_static']

html_theme_options = {
    'navbar_start': ['navbar-logo'],
    'navbar_center': ['navbar-nav'],
    'navbar_end': ['navbar-icon-links'],
    'navbar_persistent': ['search-button']
}


from uuid import uuid4 as generate_uuid
from typing import Optional
from sqlalchemy import Engine, MetaData, Table, Column
from sqlalchemy import types, select, func

def dataset_schema_size(name: str, engine: Engine, schema_name: Optional[str]=None) -> tuple[dict, dict, Optional[dict]]:
    """Return a (dataset, schema) pair or (dataset, schema, size) triplet """
    md = MetaData()
    md.reflect(engine, schema=schema_name or name)
    ds = dataset(name, md)
    return (
        ds,
        schema(name, md, ds),
        size(name, engine, ds, schema_name)
    )

def dataset(name: str, metadata: MetaData) -> dict:
    """Return a (dataset, schema) pair or (dataset, schema, size) triplet """
    return {
        '@type': 'sarus_data_spec/sarus_data_spec.Dataset',
        'uuid': generate_uuid().hex,
        'name': 'Transformed',
        'spec': {
            'transformed': {
                'transform': generate_uuid().hex,
                'arguments': [],
                'named_arguments': {},
            }
        },
        'properties': {},
        'doc': 'This ia a demo dataset for testing purpose',
    }

def schema(name: str, metadata: MetaData, dataset: dict) -> dict:
    return {
        '@type': 'sarus_data_spec/sarus_data_spec.Schema',
        'uuid': generate_uuid().hex,
        'dataset': dataset['uuid'],
        'name': name,
        'type': {
            # Slugname
            'name': name.lower(),
            'struct': {
                'fields': [
                    {
                        'name': 'sarus_data',
                        'type': {
                            'name': 'Union',
                            'union': {
                                "fields": [table(metadata.tables[name]) for name in metadata.tables]
                            },
                            'properties': {
                                'public_fields': '[]'
                            },
                        }
                    },
                    {
                        'name': 'sarus_weights',
                        'type': {
                            'name': 'Integer',
                            'integer': {
                                'min': '-9223372036854775808',
                                'max': '9223372036854775807',
                                'base': 'INT64',
                                'possible_values': []
                            },
                            'properties': {}
                        }
                    },
                    {
                        'name': 'sarus_is_public',
                        'type': {
                            'name': 'Boolean',
                            'boolean': {},
                            'properties': {}
                        }
                    },
                    {
                        'name': 'sarus_protected_entity',
                        'type': {
                            'name': 'Id',
                            'id': {
                                'base': 'STRING',
                                'unique': False
                            },
                            'properties': {}
                        }
                    },
                ],
            },
            'properties': {}
        },
        'protected': {
            'label': 'data',
            'paths': [],
            'properties': {}
        },
        'properties': {
            'max_max_multiplicity': '1',
            'foreign_keys': '',
            'primary_keys': '',
        }
    }

def table(tab: Table) -> dict:
    return {
        'name': tab.name,
        'type': {
            'name': 'Struct',
            'struct': {
                'fields': [column(col) for col in tab.columns]
            },
            'properties': {},
        }
    }

def column(col: Column) -> dict:
    if isinstance(col.type, types.Integer) or isinstance(col.type, types.BigInteger):
        return {
            'name': col.name,
            'type': {
                'name': 'Integer',
                'integer': {
                    'base': 'INT64',
                    'min': '-9223372036854775808',
                    'max': '9223372036854775807',
                    'possible_values': []
                },
                'properties': {},
            }
        }
    elif isinstance(col.type, types.Float) or isinstance(col.type, types.Numeric):
        return {
            'name': col.name,
            'type': {
                'name': 'Float64',
                'float': {
                    'base': 'FLOAT64',
                    'min': '-1.7976931348623157e+308',
                    'max': '1.7976931348623157e+308',
                    'possible_values': []
                },
                'properties': {},
            }
        }
    elif isinstance(col.type, types.String) or isinstance(col.type, types.Text) or isinstance(col.type, types.Unicode) or isinstance(col.type, types.UnicodeText):
        return {
            'name': col.name,
            'type': {
                'name': 'Text UTF-8',
                'text': {
                    'encoding': 'UTF-8'
                },
                'properties': {},
            }
        }
    elif isinstance(col.type, types.Boolean):
        return {
            'name': col.name,
            'type': {
                'name': 'Boolean',
                'boolean': {},
                'properties': {},
            }
        }
    elif isinstance(col.type, types.Date) or isinstance(col.type, types.DateTime) or isinstance(col.type, types.Time):
        return {
            'name': col.name,
            'type': {
                'name': 'Datetime',
                'datetime': {
                    'format': '%Y-%m-%d %H:%M:%S',
                    'min': '01-01-01 00:00:00',
                    'max': '9999-12-31 00:00:00',
                },
                'properties': {},
            }
        }
    else:
        return {
            'name': col.name,
            'type': {
                'name': 'Type',
                'type': {},
                'properties': {},
            }
        }

def size(name: str, engine: Engine, dataset: dict, schema_name: Optional[str]=None) -> dict:
    md = MetaData()
    md.reflect(engine, schema=schema_name or name)
    return {
        '@type': 'sarus_data_spec/sarus_data_spec.Size',
        'uuid': generate_uuid().hex,
        'dataset': dataset['uuid'],
        'name': f'{name}_sizes',
        'statistics': {
            'name': 'Union',
            'union': {
                "fields": [table_size(engine, md.tables[name]) for name in md.tables]
            },
            'properties': {}
        },
        'properties': {}
    }

def table_size(engine: Engine, tab: Table) -> dict:
    with engine.connect() as conn:
        result = conn.execute(select(func.count()).select_from(tab))
        size = result.scalar()
    multiplicity = 1.0
    return {
        'name': tab.name,
        'statistics': {
            'name': 'Struct',
            'size': str(size),
            'multiplicity': multiplicity,
            'struct': {
                'fields': [column_size(col, size, multiplicity) for col in tab.columns]
            },
            'properties': {},
        }
    }

def column_size(col: Column, size: int, multiplicity: float) -> dict:
    if isinstance(col.type, types.Integer) or isinstance(col.type, types.BigInteger):
        return {
            'name': col.name,
            'statistics': {
                'name': 'Integer',
                'size': str(size),
                'multiplicity': multiplicity,
                'integer': {
                    'base': 'INT64',
                    'min': '-9223372036854775808',
                    'max': '9223372036854775807',
                    'possible_values': []
                },
                'properties': {},
            }
        }
    elif isinstance(col.type, types.Float) or isinstance(col.type, types.Numeric):
        return {
            'name': col.name,
            'statistics': {
                'name': 'Float64',
                'size': str(size),
                'multiplicity': multiplicity,
                'float': {
                    'base': 'FLOAT64',
                    'min': '-1.7976931348623157e+308',
                    'max': '1.7976931348623157e+308',
                    'possible_values': []
                },
                'properties': {},
            }
        }
    elif isinstance(col.type, types.String) or isinstance(col.type, types.Text) or isinstance(col.type, types.Unicode) or isinstance(col.type, types.UnicodeText):
        return {
            'name': col.name,
            'statistics': {
                'name': 'Text UTF-8',
                'size': str(size),
                'multiplicity': multiplicity,
                'text': {
                    'encoding': 'UTF-8'
                },
                'properties': {},
            }
        }
    elif isinstance(col.type, types.Boolean):
        return {
            'name': col.name,
            'statistics': {
                'name': 'Boolean',
                'size': str(size),
                'multiplicity': multiplicity,
                'boolean': {},
                'properties': {},
            }
        }
    elif isinstance(col.type, types.Date) or isinstance(col.type, types.DateTime) or isinstance(col.type, types.Time):
        return {
            'name': col.name,
            'statistics': {
                'name': 'Datetime',
                'size': str(size),
                'multiplicity': multiplicity,
                'datetime': {
                    'format': '%Y-%m-%d %H:%M:%S',
                    'min': '01-01-01 00:00:00',
                    'max': '9999-12-31 00:00:00',
                },
                'properties': {},
            }
        }
    else:
        return {
            'name': col.name,
            'statistics': {
                'name': 'Type',
                'size': str(size),
                'multiplicity': multiplicity,
                'type': {},
                'properties': {},
            }
        }


from sqlalchemy import Engine, Connection, text
import typing as t
import os
from termcolor import colored

import pyqrlew as qrl

DIRNAME = os.path.join(os.getcwd(), os.path.dirname(__file__))
FILENAME = os.path.join(DIRNAME, 'queries/valid_queries.sql')

def execute_query(
    connection: Connection,
    query: str
) -> t.List[t.Dict[str, t.Any]]:
    connection.execute(text("SET search_path TO extract"))
    str_query = text(query)
    res = connection.execute(str_query)
    names = res.keys()
    return [
        {name:col for name, col in zip(names, row)}
        for row in res
    ]


def check_query(
    dataset: qrl.Dataset,
    connection: Connection,
    query: str
) -> None:
    result = execute_query(connection, query)
    rel = dataset.sql(query)
    new_query = rel.render()
    print(f"{query}->{new_query}")
    try:
        result2 = execute_query(connection, new_query)
    except TypeError as e:
        print(
            f"Sending {new_query} \nfailed with error:\n{e}"
        )
        return
    assert result == result2
    return


def test_results(database):
    engine = database.engine()
    with engine.connect() as conn:
        with open(FILENAME, 'r') as f:
            for query in f:
                if not query.startswith('--'):
                    print('\n\n', colored(query, 'blue'))
                    check_query(database.extract(), conn, query)



from importlib.resources import files
import logging
import pyqrlew as qrl
from pyqrlew.io import PostgreSQL

logging.basicConfig(level=logging.DEBUG)

db = PostgreSQL()

ds = db.extract()
# ds = db.financial()
# ds = db.imdb()
# ds = db.hepatitis()

print(ds.relations())
print(ds.sql("select * from census").dot())



import shutil
from pathlib import Path
import importlib.resources as pkg_resources
import json
import logging
from sqlalchemy import MetaData, Engine, exists, select, column, table, text
from datasets import sources
from datasets.databases import PostgreSQL as EmptyPostgreSQL
from pyqrlew.io.database import dataset_schema_size
import pyqrlew as qrl

NAME: str = 'pyqrlew-db'
USER: str = 'postgres'
PASSWORD: str = 'pyqrlew-db'
PORT: str = 5433

class PostgreSQL(EmptyPostgreSQL):
    def __init__(self) -> None:
        super().__init__(NAME, USER, PASSWORD, PORT)
    
    def load_resource(self, schema: str, src: Path) -> 'PostgreSQL':
        with self.engine().connect() as conn:
            res = list(conn.execute(select(column('schema_name')).select_from(table('schemata', schema='information_schema')).where(column('schema_name') == schema)))
            schema_exists = len(res)==1
        if not schema_exists:
            dst = Path('/tmp') / schema
            shutil.copyfile(src, dst)
            self.load(dst)
        return self
    
    def load_extract(self) -> 'PostgreSQL':
        return self.load_resource('extract', pkg_resources.files(sources) / 'extract' / 'extract.sql')

    def load_financial(self) -> 'PostgreSQL':
        return self.load_resource('financial', pkg_resources.files(sources) / 'financial' / 'financial.sql')

    def load_hepatitis(self) -> 'PostgreSQL':
        return self.load_resource('hepatitis_std', pkg_resources.files(sources) / 'hepatitis' / 'Hepatitis_std.sql')

    def load_imdb(self) -> 'PostgreSQL':
        return self.load_resource('imdb_ijs', pkg_resources.files(sources) / 'imdb' / 'imdb_ijs.sql')
    
    def extract(self) -> qrl.Dataset:
        self.load_extract()
        dataset, schema, size = dataset_schema_size('extract', self.engine())
        return qrl.Dataset(json.dumps(dataset), json.dumps(schema), json.dumps(size))
    
    def financial(self) -> qrl.Dataset:
        self.load_financial()
        dataset, schema, size = dataset_schema_size('financial', self.engine())
        return qrl.Dataset(json.dumps(dataset), json.dumps(schema), json.dumps(size))
    
    def hepatitis(self) -> qrl.Dataset:
        self.load_hepatitis()
        dataset, schema, size = dataset_schema_size('hepatitis_std', self.engine())
        return qrl.Dataset(json.dumps(dataset), json.dumps(schema), json.dumps(size))

    def imdb(self) -> qrl.Dataset:
        self.load_imdb()
        dataset, schema, size = dataset_schema_size('imdb_ijs', self.engine())
        return qrl.Dataset(json.dumps(dataset), json.dumps(schema), json.dumps(size))
